/* GtkZephyr - a GTK+ based front end to MIT's Zephyr Notification Service
 * Copyright (C) 2000 Sean Dilda
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */


#include <sys/param.h>
#include <zephyr/zephyr.h>
#include <com_err.h>
#include <glib.h>
#include <stdio.h>
#include <stdlib.h>
#include <pwd.h>
#include <unistd.h>
#include <netdb.h>
#include "gtkzephyr.h"
#include "userdialog.h"
#include "chatdialog.h"
#include "zcalls.h"
#include "prefs.h"
#include "errorbox.h"

/*  Function types that /should/ be in zephyr/zephyr.h, but aren't */
Code_t ZUnsetLocation();
Code_t ZSetLocation(char *exposure);


typedef struct {
    ZUnique_Id_t uid;
    handlerfunc handler;
    gtkzephyr_subscrip *subscrip;
} wait_node;

GHashTable *waittable;
static gchar* local_userid;

#define try_zcall(func, error)  if ((retval = func) != ZERR_NONE) { \
                                  gtkzephyr_error_with_code(error, retval); \
                                  return; \
                                }

#define try_zcall2(func, error)  if ((retval = func) != ZERR_NONE) { \
                                  gtkzephyr_error_with_code(error, retval); \
                                  return TRUE; \
                                }

guint zuid_hash(ZUnique_Id_t *uid) {
  /* rather cheap.. just index it based on the UNIX timestamp */
  return uid->tv.tv_sec;
}



static gchar *get_exposure_level() {

  gchar *exposure;

  exposure = ZGetVariable("exposure");
  DEBUG_INFO("Read exposure level: %s\n", exposure);
  if (!g_strcasecmp(exposure, EXPOSE_NONE)) {
    return EXPOSE_NONE;
  } else if (!g_strcasecmp(exposure, EXPOSE_OPSTAFF)) {
    return EXPOSE_OPSTAFF;
  } else if (!g_strcasecmp(exposure, EXPOSE_REALMANN)) {
    return EXPOSE_REALMANN;
  } else if (!g_strcasecmp(exposure, EXPOSE_NETVIS)) {
    return EXPOSE_NETVIS;
  } else if (!g_strcasecmp(exposure, EXPOSE_NETANN)) {
    return EXPOSE_NETANN;
  } else {
    return EXPOSE_REALMVIS;
  }
}



void gtkzephyr_connect() {

  Code_t retval;
  ZSubscription_t sub;

  waittable = g_hash_table_new((GHashFunc) zuid_hash, (GCompareFunc) ZCompareUID);
  try_zcall(ZInitialize(),"while initializing zephyrs");
  try_zcall(ZOpenPort(NULL), "while opening port");
  try_zcall(ZSetLocation(get_exposure_level()), "while setting location");
  
  /* FIXME: This *should* be a for loop going through $HOME/.zephyr.subs but for
   * now we'll just subscribe to MESSAGE,PERSONAL,%me% */

  sub.zsub_class = "MESSAGE";
  sub.zsub_classinst = "PERSONAL";
  sub.zsub_recipient = ZGetSender();

  try_zcall(ZSubscribeTo(&sub, 1, 0), "Error while subscribing");

}



void gtkzephyr_sub_login_class(gtkzephyr_user *user) {

  Code_t retval;
  ZSubscription_t sub;

  sub.zsub_class = LOGIN_CLASS;
  sub.zsub_classinst = user->userrealm;
  sub.zsub_recipient = "";

  try_zcall(ZSubscribeTo(&sub, 1, 0), "Error while subscribing");

}



void gtkzephyr_sub(gchar *class, gchar *instance, gchar *recipient) {

  Code_t retval;
  ZSubscription_t sub;

  sub.zsub_class = class;
  sub.zsub_classinst = instance;
  sub.zsub_recipient = recipient;

  try_zcall(ZSubscribeTo(&sub, 1, 0), "Error while subscribing");

}
  



void gtkzephyr_unsub(gchar *class, gchar *instance, gchar *recipient) {
  // STUB -- FIXME!!
}



void gtkzephyr_punt(gchar *class, gchar *instance, gchar *recipient) {
  // STUB --FIXME!!
}




void gtkzephyr_disconnect() {

  Code_t retval;

  try_zcall(ZCancelSubscriptions(0), "Error while canceling subscriptions");
  try_zcall(ZUnsetLocation(), "Error while unsetting location");
  try_zcall(ZClosePort(), "Error while closing port");

}



void update_locs(gpointer userpointer, gpointer foo) {

  Code_t retval;
  int numlocs;
  char name[BUFSIZ];
  register char *cp;
  gtkzephyr_user *user = (gtkzephyr_user *) userpointer;

  strncpy(name,user->userid,BUFSIZ - 1);
  if (!strchr(name,'@')) {
    gint namelen = strlen(name);
    cp = name + namelen;
    *cp++ = '@';
    strncpy(cp,ZGetRealm(), BUFSIZ - namelen - 1);
  }
  name[BUFSIZ - 1] = '\0';
  try_zcall(ZLocateUser(name, &numlocs, ZAUTH), "locating user");
  if (numlocs) {
    user->status = 1;
  } else {
    user->status = 0;
  }

}



void gtkzephyr_update_userlist(GList *userlist) {

  g_list_foreach(userlist, (GFunc) update_locs, NULL);

}




gchar *gtkzephyr_get_signature(gchar *message) {

  return message;

}



gchar *gtkzephyr_get_body(gchar *message, gint mlength) {

  gchar *ptr;

  ptr = message + strlen(message) + 1;
  mlength = mlength - (ptr - message);

  if (mlength <= 0) {
    DEBUG_INFO("Returning NULL.. no body in message %d - %d = %d\n", mlength, ptr - message, mlength - (ptr - message));
    return (gchar *) NULL;
  }
  ptr = g_strndup(ptr, mlength);

  return g_strchomp(ptr);
  
}



void gtkzephyr_handle_unknown_notice(ZNotice_t *notice) {

  g_print("** New Notice **\n");
  g_print("z_packet: %s\n", notice->z_packet);
  g_print("z_version: %s\n", notice->z_version);
  g_print("z_kind: ");
  switch (notice->z_kind) {
    case UNSAFE: g_print("UNSAFE"); break;
    case UNACKED:  g_print("UNACKED"); break;
    case ACKED: g_print("ACKED"); break;
    case HMACK: g_print("HMACK"); break;
    case HMCTL: g_print("HMCTL"); break;
    case SERVACK: g_print("SERVACK"); break;
    case SERVNAK: g_print("SERVNAK"); break;
    case CLIENTACK: g_print("CLIENTACK"); break;
    case STAT: g_print("STAT"); break;
  }
  g_print("\n");
  g_print("z_uid: %d:%d.%d\n", notice->z_uid.zuid_addr.s_addr,
      (gint) notice->z_uid.tv.tv_sec, (gint) notice->z_uid.tv.tv_usec);
  g_print("z_time: %d->%d\n", (gint) notice->z_time.tv_sec,
      (gint) notice->z_time.tv_usec);
  g_print("z_port: %d\n", notice->z_port);
  g_print("z_auth: %d\n", notice->z_auth);
  g_print("z_authent_len: %d\n", notice->z_authent_len);
  g_print("z_ascii_authent: %s\n", notice->z_ascii_authent);
  g_print("z_class: %s\n", notice->z_class);
  g_print("z_class_inst: %s\n", notice->z_class_inst);
  g_print("z_opcode: %s\n", notice->z_opcode);
  g_print("z_sender: %s\n", notice->z_sender);
  g_print("z_recipient: %s\n", notice->z_recipient);
  g_print("z_default_format; %s\n", notice->z_default_format);
  g_print("z_multinotice: %s\n", notice->z_multinotice);
  g_print("z_multiuid: %d:%d.%d\n", notice->z_multiuid.zuid_addr.s_addr,
      (gint) notice->z_multiuid.tv.tv_sec,
      (gint) notice->z_multiuid.tv.tv_usec);
  g_print("z_checksum: %d\n", (gint) notice->z_checksum);
  g_print("z_num_other_files: %d\n", notice->z_num_other_fields);
  g_print("z_message_len: %d\n", notice->z_message_len);
  g_print("z_message: %s\n", gtkzephyr_get_body(notice->z_message,
	notice->z_message_len));


}



void handle_message(ZNotice_t *notice) {
  /* We're going to have to do this as an either or..  if the chatwindow is
   * open, we just use that, otherwise we set it up for the user dialog */

  if (!strcmp(notice->z_opcode,"PING")) {
    // ignore PING's for now..
    ZFreeNotice(notice);
  } else if (!strcmp(notice->z_class, LOGIN_CLASS)) {

    gtkzephyr_user *user = gtkzephyr_get_user(notice->z_sender);

#ifdef DEBUG
    gtkzephyr_handle_unknown_notice(notice);
#endif

    if (strcmp(notice->z_sender, notice->z_class_inst)) {
      gtkzephyr_error("Forged LOGIN_CLASS!\n");
      ZFreeNotice(notice);
      return;
    }

    if (!strcmp(notice->z_opcode, LOGIN_USER_LOGIN)) {
      DEBUG_INFO("%s logging in..\n", user->userid);
      if (!user->status) {
	/* if this is the only login for user */
	DEBUG_INFO("Woah! It's a real login..\n");
	gtkzephyr_handle_user_login(user);
      }
    } else if (!strcmp(notice->z_opcode, LOGIN_USER_FLUSH)) {
      /* This is apparently never used, but is in zephyr.h, so we'll
       * include it anyways */
      DEBUG_INFO("%s flushing locs\n", user->userid);
      gtkzephyr_handle_user_logout(user);
    } else if (!strcmp(notice->z_opcode, LOGIN_USER_LOGOUT)) {
      DEBUG_INFO("%s logging out..\n", user->userid);
      update_locs(user, NULL);
      if (!user->status) {
      	DEBUG_INFO("Woah! It's a real log out\n");
	gtkzephyr_handle_user_logout(user);
      }
    } else
      g_assert_not_reached();

    ZFreeNotice(notice);

  } else if (!strcmp(notice->z_class, DEFAULT_CLASS) && !strcmp(notice->z_class_inst, DEFAULT_INSTANCE)) {
    
    gtkzephyr_user *user = gtkzephyr_get_user(notice->z_sender);

    if (!user) {
      gchar *userid = g_strndup(notice->z_sender,
	  strchr(notice->z_sender, '@') - notice->z_sender);

      if (prefstruct->add_new_users) {
	gtkzephyr_add_user(userid);
      } else {
	gtkzephyr_make_user(userid);
      }
      // Hrm.. maybe the above two functions should just return a struct
      user = gtkzephyr_get_user(notice->z_sender);
      g_free(userid);
    }
    if (GTKZ_SUBSCRIP(user)->chatwindow && GTK_WIDGET_VISIBLE(GTK_WIDGET(GTKZ_SUBSCRIP(user)->chatwindow))) {
      gtkzephyr_chat_handle_message(notice, GTKZ_SUBSCRIP(user));
    } else {
      gtkzephyr_user_dialog_handle_message(notice, GTKZ_SUBSCRIP(user));
    }
  } else {

    gtkzephyr_inst *inst = gtkzephyr_get_inst(notice->z_class, notice->z_class_inst, notice->z_recipient);

    if (!inst) {
      inst = gtkzephyr_make_inst(notice->z_class, notice->z_class_inst, notice->z_recipient);
    }
    if (GTKZ_SUBSCRIP(inst)->chatwindow && GTK_WIDGET_VISIBLE(GTK_WIDGET(GTKZ_SUBSCRIP(inst)->chatwindow))) {
      gtkzephyr_chat_handle_message(notice, GTKZ_SUBSCRIP(inst));
    } else {
      gtkzephyr_user_dialog_handle_message(notice, GTKZ_SUBSCRIP(inst));
    }
    
  }

}



void handle_servack(ZNotice_t *notice) {

  wait_node *node = (wait_node*) g_hash_table_lookup(waittable, &notice->z_uid);

  if (node) {
    g_hash_table_remove(waittable, &notice->z_uid);
    (*node->handler) (notice, node->subscrip);
    free(node);
  } else if (!strcmp(notice->z_opcode, LOGIN_USER_LOGIN)) {
    /* just acknowledging your login, ignore */
  } else {
    gtkzephyr_handle_unknown_notice(notice);
  }
  ZFreeNotice(notice);

}



gint gtkzephyr_checknewnotices(gpointer a) {

  ZNotice_t *notice;
  struct sockaddr_in from;
  Code_t retval;

  while (ZPending()) {

    notice = g_new(ZNotice_t, 1);
    try_zcall2(ZReceiveNotice(notice, &from), "while receiving notice");

    switch (notice->z_kind) {
      case UNSAFE:
      case UNACKED:
      case ACKED:
        handle_message(notice);
        break;
      case SERVACK:
      case SERVNAK:
        handle_servack(notice);
        break;
      default:
        gtkzephyr_handle_unknown_notice(notice);
	ZFreeNotice(notice);
    }

  }
  return TRUE;

}




void gtkzephyr_flush_loc(GtkWidget *widget) {

  int retval;

  try_zcall(ZFlushMyLocations(), "while flushing locations");

}





void gtkzephyr_hide(GtkWidget *widget) {

  Code_t retval;

  try_zcall(ZSetLocation(EXPOSE_OPSTAFF),"while hiding.");

}




void gtkzephyr_unhide(GtkWidget *widget) {

  Code_t retval;

  try_zcall(ZSetLocation(get_exposure_level()),"while unhiding.");

}



void gtkzephyr_send_text(gtkzephyr_subscrip *subscrip, GtkEditable *text, handlerfunc handle_function) {

  ZNotice_t notice;
  gchar *message, *buffer, *class, *inst, *opcode, *signature, *recipient;
  wait_node *node;
  gint msgsize;

  if (subscrip->type == GTKZ_USER_TYPE) {
    class = (buffer = ZGetVariable("zwrite-class")) ? g_strdup(buffer) : NULL;
    inst = (buffer = ZGetVariable("zwrite-inst")) ? g_strdup(buffer) : NULL;
    recipient = GTKZ_USER(subscrip)->userrealm;
  } else {
    class = g_strdup(GTKZ_INST(subscrip)->class);
    inst = g_strdup(GTKZ_INST(subscrip)->instance);
    recipient = "";
  }

  opcode = (buffer = ZGetVariable("zwrite-opcode")) ? g_strdup(buffer) : NULL;
  buffer = ZGetVariable("zwrite-signature");
  if (buffer) {
    signature = g_strdup(buffer);
  } else {
    signature = g_strdup(g_get_real_name());
  }

  msgsize = 0;
  message = NULL;
  if (signature) {
    message = malloc(strlen(signature) + 2);
    strcpy(message, signature);
    msgsize = strlen(message);
    message[msgsize++] = '\0';
  } else {
    message = malloc(1);
    message[msgsize++] = '\0';
  }
  buffer = gtk_editable_get_chars(text, 0, -1);
  message = realloc(message, (unsigned)(msgsize + strlen(buffer) + 1));
  strcpy(message+msgsize, buffer);
  msgsize += strlen(buffer);
  g_free(buffer);


  bzero((char *)&notice, sizeof(notice));
  notice.z_kind = ACKED;
  notice.z_port = 0;
  notice.z_opcode = opcode ? opcode : "";
  notice.z_class = class ? class : "MESSAGE";
  notice.z_class_inst = inst ? inst : "PERSONAL";
  notice.z_sender = 0;  /* libzephyr will fill in later */
  notice.z_recipient = recipient;
  notice.z_default_format = signature ?
    "Class $class, Instance $instance:\nTo: @bold($recipient) at $time $date\nFrom: @bold($1) <$sender>\n\n$2" :
    "Class $class, Instance $instance:\nTo: @bold($recipient) at $time $date\n$message";
  notice.z_message_len = msgsize;
  notice.z_message = message;

  node = g_new(wait_node, 1);
  node->subscrip = subscrip;
  node->handler = handle_function;
  ZSendNotice(&notice, ZAUTH);
  node->uid = notice.z_uid;;
  g_hash_table_insert(waittable, &node->uid, node);

  g_free(message);
  g_free(class);
  g_free(inst);
  g_free(opcode);

}

gchar *gtkzephyr_get_local_userid() {

  if (!local_userid) {
    gchar *fulluserid = ZGetSender();
    local_userid = g_strndup(fulluserid, strchr(fulluserid, '@') - fulluserid);
  }

  return local_userid;

}

void gtkzephyr_macro_sub(gchar *field) {

  static int initedhosts = 0;
  static gchar ourhost[MAXHOSTNAMELEN], ourhostcanon[MAXHOSTNAMELEN];

  if (!initedhosts) {
    
    struct hostent *hent;

    if (gethostname(ourhost, sizeof(ourhost) - 1) == -1) {
      gtkzephyr_error("Unambe to retreive hostname, some zephyr subscriptions might be wrong.");
    } else if (!(hent = gethostbyname(ourhost))) {
      gtkzephyr_error("Unable to resolve hostname, some subscriptions might be wrong.");
    }

    strncpy(ourhostcanon, hent->h_name, sizeof(ourhostcanon) - 1);
    initedhosts = 1;
  }
  if (!strcmp(field, TOKEN_ME))
    strncpy(field, ZGetSender(), BUFSIZ);
  else if (!strcmp(field, TOKEN_HOSTNAME))
    strncpy(field, ourhost, BUFSIZ);
  else if (!strcmp(field, TOKEN_CANONNAME))
    strncpy(field, ourhostcanon, BUFSIZ);
}
